@@ -22,15 +22,19 @@ module Codec.Zlib
     ( -- * Inflate
       Inflate
     , initInflate
+    , initInflateWithDictionary
     , withInflateInput
     , finishInflate
+    , flushInflate
       -- * Deflate
     , Deflate
     , initDeflate
+    , initDeflateWithDictionary
     , withDeflateInput
     , finishDeflate
+    , flushDeflate
       -- * Data types
-    , WindowBits (WindowBits)
+    , WindowBits (..)
     , defaultWindowBits
     , ZlibException (..)
     ) where
@@ -39,7 +43,7 @@ import Codec.Zlib.Lowlevel
 import Foreign.ForeignPtr
 import Foreign.C.Types
 import Data.ByteString.Unsafe
-import Codec.Compression.Zlib (WindowBits (WindowBits), defaultWindowBits)
+import Codec.Compression.Zlib (WindowBits(WindowBits), defaultWindowBits)
 import qualified Data.ByteString as S
 import Data.ByteString.Lazy.Internal (defaultChunkSize)
 import Control.Monad (when)
@@ -50,7 +54,8 @@ type ZStreamPair = (ForeignPtr ZStreamStruct, ForeignPtr CChar)
 
 -- | The state of an inflation (eg, decompression) process. All allocated
 -- memory is automatically reclaimed by the garbage collector.
-newtype Inflate = Inflate ZStreamPair
+-- Also can contain the inflation dictionary that is used for decompression.
+newtype Inflate = Inflate (ZStreamPair, Maybe S.ByteString)
 
 -- | The state of a deflation (eg, compression) process. All allocated memory
 -- is automatically reclaimed by the garbage collector.
@@ -77,10 +82,18 @@ newtype Deflate = Deflate ZStreamPair
 -- * #define Z_BUF_ERROR    (-5)
 --
 -- * #define Z_VERSION_ERROR (-6)
+
 data ZlibException = ZlibException Int
     deriving (Show, Typeable)
 instance Exception ZlibException
 
+-- | Some constants for the error codes, used internally
+zNeedDict :: CInt
+zNeedDict = 2
+
+zBufError :: CInt
+zBufError = -5
+
 -- | Initialize an inflation process with the given 'WindowBits'. You will need
 -- to call 'withInflateInput' to feed compressed data to this and
 -- 'finishInflate' to extract the final chunk of decompressed data.
@@ -92,7 +105,21 @@ initInflate w = do
     fbuff <- mallocForeignPtrBytes defaultChunkSize
     withForeignPtr fbuff $ \buff ->
         c_set_avail_out zstr buff $ fromIntegral defaultChunkSize
-    return $ Inflate (fzstr, fbuff)
+    return $ Inflate ((fzstr, fbuff), Nothing)
+
+-- | Initialize an inflation process with the given 'WindowBits'. 
+-- Unlike initInflate a dictionary for inflation is set which must
+-- match the one set during compression.
+initInflateWithDictionary :: WindowBits -> S.ByteString -> IO Inflate
+initInflateWithDictionary w bs = do
+    zstr <- zstreamNew
+    inflateInit2 zstr w
+    fzstr <- newForeignPtr c_free_z_stream_inflate zstr
+    fbuff <- mallocForeignPtrBytes defaultChunkSize
+
+    withForeignPtr fbuff $ \buff ->
+        c_set_avail_out zstr buff $ fromIntegral defaultChunkSize
+    return $ Inflate ((fzstr, fbuff), Just bs)
 
 -- | Initialize a deflation process with the given compression level and
 -- 'WindowBits'. You will need to call 'withDeflateInput' to feed uncompressed
@@ -109,6 +136,25 @@ initDeflate level w = do
         c_set_avail_out zstr buff $ fromIntegral defaultChunkSize
     return $ Deflate (fzstr, fbuff)
 
+-- | Initialize an deflation process with the given compression level and
+-- 'WindowBits'.
+-- Unlike initDeflate a dictionary for deflation is set.
+initDeflateWithDictionary :: Int -- ^ Compression level
+			  -> S.ByteString -- ^ Deflate dictionary
+			  -> WindowBits -> IO Deflate
+initDeflateWithDictionary level bs w = do
+    zstr <- zstreamNew
+    deflateInit2 zstr level w 8 StrategyDefault
+    fzstr <- newForeignPtr c_free_z_stream_deflate zstr
+    fbuff <- mallocForeignPtrBytes defaultChunkSize
+
+    unsafeUseAsCStringLen bs $ \(cstr, len) -> do
+        c_call_deflate_set_dictionary zstr cstr $ fromIntegral len
+
+    withForeignPtr fbuff $ \buff ->
+        c_set_avail_out zstr buff $ fromIntegral defaultChunkSize
+    return $ Deflate (fzstr, fbuff)
+
 -- | Feed the given 'S.ByteString' to the inflater. This function takes a
 -- function argument which takes a \"popper\". A popper is an IO action that
 -- will return the next bit of inflated data, returning 'Nothing' when there is
@@ -122,11 +168,21 @@ initDeflate level w = do
 withInflateInput
     :: Inflate -> S.ByteString -> (IO (Maybe S.ByteString) -> IO a)
     -> IO a
-withInflateInput (Inflate (fzstr, fbuff)) bs f =
+withInflateInput (Inflate ((fzstr, fbuff), inflateDictionary)) bs f =
     withForeignPtr fzstr $ \zstr ->
         unsafeUseAsCStringLen bs $ \(cstr, len) -> do
             c_set_avail_in zstr cstr $ fromIntegral len
-            f $ drain fbuff zstr c_call_inflate_noflush False
+            f $ drain fbuff zstr inflate False
+  where
+    inflate zstr = do
+	res <- c_call_inflate_noflush zstr
+	if (res == zNeedDict)
+	    then maybe (throwIO $ ZlibException $ fromIntegral zNeedDict) -- no dictionary supplied so throw error
+		       (\dict -> (unsafeUseAsCStringLen dict $ \(cstr, len) -> do
+				    c_call_inflate_set_dictionary zstr cstr $ fromIntegral len
+				    c_call_inflate_noflush zstr))
+		       inflateDictionary
+	    else return res
 
 drain :: ForeignPtr CChar -> ZStream' -> (ZStream' -> IO CInt) -> Bool
       -> IO (Maybe S.ByteString)
@@ -136,7 +192,7 @@ drain fbuff zstr func isFinish = do
         then return Nothing
         else withForeignPtr fbuff $ \buff -> do
             res <- func zstr
-            when (res < 0 && res /= (-5))
+            when (res < 0 && res /= zBufError)
                 $ throwIO $ ZlibException $ fromIntegral res
             avail <- c_get_avail_out zstr
             let size = defaultChunkSize - fromIntegral avail
@@ -149,17 +205,30 @@ drain fbuff zstr func isFinish = do
                     return $ Just bs
                 else return Nothing
 
+
 -- | As explained in 'withInflateInput', inflation buffers your decompressed
 -- data. After you call 'withInflateInput' with your last chunk of compressed
 -- data, you will likely have some data still sitting in the buffer. This
 -- function will return it to you.
 finishInflate :: Inflate -> IO S.ByteString
-finishInflate (Inflate (fzstr, fbuff)) =
+finishInflate (Inflate ((fzstr, fbuff), _)) =
     withForeignPtr fzstr $ \zstr ->
         withForeignPtr fbuff $ \buff -> do
             avail <- c_get_avail_out zstr
             let size = defaultChunkSize - fromIntegral avail
-            S.packCStringLen (buff, size)
+            bs <- S.packCStringLen (buff, size)
+            c_set_avail_out zstr buff
+                $ fromIntegral defaultChunkSize
+            return bs
+
+-- | Flush the inflation buffer. Useful for interactive application.
+--
+-- This is actually a synonym for 'finishInflate'. It is provided for its more
+-- semantic name.
+--
+-- Since 0.0.3
+flushInflate :: Inflate -> IO S.ByteString
+flushInflate = finishInflate
 
 -- | Feed the given 'S.ByteString' to the deflater. This function takes a
 -- function argument which takes a \"popper\". A popper is an IO action that
@@ -188,3 +257,13 @@ finishDeflate :: Deflate -> (IO (Maybe S.ByteString) -> IO a) -> IO a
 finishDeflate (Deflate (fzstr, fbuff)) f =
     withForeignPtr fzstr $ \zstr ->
         f $ drain fbuff zstr c_call_deflate_finish True
+
+-- | Flush the deflation buffer. Useful for interactive application.
+--
+-- Internally this passes Z_SYNC_FLUSH to the zlib library.
+--
+-- Since 0.0.3
+flushDeflate :: Deflate -> (IO (Maybe S.ByteString) -> IO a) -> IO a
+flushDeflate (Deflate (fzstr, fbuff)) f =
+    withForeignPtr fzstr $ \zstr ->
+        f $ drain fbuff zstr c_call_deflate_flush True
