@@ -27,6 +27,16 @@ int deflate_init2(z_stream *stream, int level, int methodBits,
 	return deflateInit2(stream, level, Z_DEFLATED, methodBits, memlevel, strategy);
 }
 
+int inflate_set_dictionary(z_stream *stream, const char* dictionary, 
+                            unsigned int dictLength) {
+        return inflateSetDictionary(stream, dictionary, dictLength);
+}
+
+int deflate_set_dictionary(z_stream *stream, const char* dictionary, 
+                            unsigned int dictLength) {
+        return deflateSetDictionary(stream, dictionary, dictLength);
+}
+
 void free_z_stream_inflate (z_stream *stream)
 {
 	inflateEnd(stream);
@@ -71,6 +81,11 @@ int call_deflate_noflush (z_stream *stream)
 	return deflate(stream, Z_NO_FLUSH);
 }
 
+int call_deflate_flush (z_stream *stream)
+{
+	return deflate(stream, Z_SYNC_FLUSH);
+}
+
 int call_deflate_finish (z_stream *stream)
 {
 	return deflate(stream, Z_FINISH);
